#ifndef GRAFO_H_
#define GRAFO_H_

#include "Lista.h"

#include "cstdlib"
#include "fstream"
#include "iostream"

/*
 * Por cuestiones de implementacin el arco se define fuera del mbito de la
 * clase Grafo; esto es necesario para que funcione correctamente la sobrecarga
 * del operador de salida de los streams (y as poder mostrar Arcos de grafos
 * parametrizados con cualquier costo; ver el operador para mostrar el grafo).
 * Sin embargo, la forma de declarar los arcos sigue siendo Grafo<Costo>::Arco.
 */
template <typename C> class Arco_;


/*
 * Implementacin del grafo utilizando listas de adyacencia.
 */

template <typename C> class Grafo
{
public:
    // Se renombra a la clase Arco_ a Arco; esto permite que las declaraciones
    // de los arcos de la forma Grafo<Costo>::Arco sigan funcionando.
    typedef Arco_<C> Arco;

public:
	Grafo();
	Grafo(const Grafo & otroGrafo);

	~Grafo();

	Grafo & operator = (const Grafo & otroGrafo);

	// Devuelve true si la cantidad de vrtices es cero
	bool estaVacio() const;
	// Indica la cantidad de vrtices del grafo
	int devolverLongitud() const;

	bool existeVertice(int vertice) const;

	bool existeArco(int origen, int destino) const;

	// PRE CONDICION: existeArco(origen, destino)
	const C & costoArco(int origen, int destino) const;

	void devolverVertices(Lista<int> & vertices) const;

	void devolverAdyacentes(int origen, Lista<Arco> & adyacentes) const;

	void agregarVertice(int vertice);

	// POST CONDICION: Para todo vrtice v != vertice: !existeArco(v, vertice) && !existeArco(vertice, v)
	void eliminarVertice(int vertice);

	// PRE CONDICION: existeArco(origen, destino)
	C & costoArco(int origen, int destino);

	// PRE CONDICION: existeVertice(origen) && existeVertice(destino)
	// POST CONDICION: existeArco(origen, destino)
	void agregarArco(int origen, int destino, const C & costo);

	// POST CONDICION: !existeArco(origen, destino)
	void eliminarArco(int origen, int destino);

	void vaciar();

private:
    class Vertice
    {
    public:
        Vertice();
        Vertice(int vertice);
        bool operator ==(const Vertice & otroVertice) const;
	public:
		Lista<Arco> adyacentes;
		int vertice;
	}; // class Vertice

	Lista<Vertice> vertices;
}; // class Grafo


/*
 * Arco_
 */

template <typename C> class Arco_
{
public:
    Arco_();
    Arco_(int adyacente);
    Arco_(int adyacente, const C & costo);
    int devolverAdyacente() const;
    const C & devolverCosto() const;
    C & devolverCosto();
    bool operator ==(const Arco_ & otroArco) const;
private:
    int vertice;
    C costo;
}; // class Arco_

template <typename C> Arco_<C>::Arco_()
{
}

template <typename C> Arco_<C>::Arco_(int adyacente)
{
    vertice = adyacente;
}

template <typename C> Arco_<C>::Arco_(int adyacente, const C & costo)
{
	this->vertice = adyacente;
	this->costo = costo;
}

template <typename C> int Arco_<C>::devolverAdyacente() const
{
	return vertice;
}

template <typename C> const C & Arco_<C>::devolverCosto() const
{
	return costo;
}

template <typename C> C & Arco_<C>::devolverCosto()
{
	return costo;
}

template <typename C> bool Arco_<C>::operator ==(const Arco_ & otroArco) const
{
    return (vertice == otroArco.vertice);
}

template <typename C> std::ostream & operator << (std::ostream & salida, const Arco_<C> & arco)
{
	salida << arco.devolverAdyacente() << " (" << arco.devolverCosto() << ")";
	return salida;
}


/*
 * Vertice
 */

template <typename C> Grafo<C>::Vertice::Vertice()
{
}

template <typename C> Grafo<C>::Vertice::Vertice(int vertice)
{
    this->vertice = vertice;
}

/*
 * Se sobrecaga el operador de salida de los streams para permitir mostrar un arco genrico.
 * Tener en cuenta que se asume que el tipo del costo soporta dicho operador.
 */

template <typename C> bool Grafo<C>::Vertice::operator ==(const Vertice & v) const
{
    return (this->vertice == v.vertice);
}


/*
 * Grafo
 */

template <typename C> Grafo<C>::Grafo()
{
}

template <typename C> Grafo<C>::~Grafo()
{
    vaciar();
}

template <typename C> Grafo<C>::Grafo(const Grafo & otroGrafo)
{
   operator=(otroGrafo);
}

template <typename C> Grafo<C> & Grafo<C>::operator =(const Grafo & otroGrafo)
{
    // Se verifica que no se intente asignar un grafo a s mismo
    if (this != &otroGrafo) {
        vertices = otroGrafo.vertices;
    }
    return *this;
}

template <typename C> bool Grafo<C>::estaVacio() const
{
    return vertices.estaVacia();
}

template <typename C> int Grafo<C>::devolverLongitud() const
{
    return vertices.devolverLongitud();
}

template <typename C> bool Grafo<C>::existeVertice(int vertice) const
{
    Vertice v(vertice);

    return !vertices.buscar(v).llegoAlFinal();
}

template <typename C> bool Grafo<C>::existeArco(int origen, int destino) const
{
    bool existe = false;

    Vertice orig(origen);
    typename Lista<typename Grafo<C>::Vertice>::ConstIterador posOrig = vertices.buscar(orig);

    if (!posOrig.llegoAlFinal()) {
        Arco dest(destino);
        if (!posOrig.elementoActual().adyacentes.buscar(dest).llegoAlFinal())
            existe = true;
    }

    return existe;
}

// PRE CONDICION: existeArco(origen, destino)
template <typename C> const C & Grafo<C>::costoArco(int origen, int destino) const
{
    assert(existeArco(origen, destino));

    Vertice orig(origen);
    typename Lista<Vertice>::ConstIterador posOrig = vertices.buscar(orig);

    if (!posOrig.llegoAlFinal()) {
        Arco dest(destino);
        typename Lista<typename Grafo<C>::Arco>::ConstIterador posDest = posOrig.elementoActual().adyacentes.buscar(dest);
        if (!posDest.llegoAlFinal())
            return posDest.elementoActual().devolverCosto();
    }
}

template <typename C> void Grafo<C>::devolverVertices(Lista<int> & resultado) const
{
    resultado.vaciar();

    typename Lista<Vertice>::ConstIterador itVert = this->vertices.devolverIterador();
    while (!itVert.llegoAlFinal())
    {
        resultado.agregarFinal(itVert.elementoActual().vertice);
        itVert.avanzar();
    }
}

template <typename C> void Grafo<C>::devolverAdyacentes(int origen, Lista<Arco> & adyacentes) const
{
    adyacentes.vaciar();

    Vertice orig(origen);
    typename Lista<Vertice>::ConstIterador posOrig = this->vertices.buscar(orig);

    if (!posOrig.llegoAlFinal()) {
        adyacentes = posOrig.elementoActual().adyacentes;
    }
}

template <typename C> void Grafo<C>::agregarVertice(int vertice)
{
    Vertice v(vertice);

    if (vertices.buscar(v).llegoAlFinal())
        vertices.agregarFinal(v);
}

// POST CONDICION: Para todo vrtice v != vertice: !existeArco(v, vertice) && !existeArco(vertice, v)
template <typename C> void Grafo<C>::eliminarVertice(int vertice)
{
    Vertice orig(vertice);
    typename Lista<Vertice>::Iterador posOrig = vertices.buscar(orig);

    if (!posOrig.llegoAlFinal())
        vertices.eliminar(posOrig);

    typename Lista<Vertice>::Iterador itVert = vertices.devolverIterador();
    while (!itVert.llegoAlFinal()) {
        eliminarArco(itVert.elementoActual().vertice, vertice);
        itVert.avanzar();
    }
}

// PRE CONDICION: existeArco(origen, destino)
template <typename C> C & Grafo<C>::costoArco(int origen, int destino)
{
    assert(existeArco(origen, destino));

    Vertice orig(origen);
    typename Lista<Vertice>::Iterador posOrig = vertices.buscar(orig);

    if (!posOrig.llegoAlFinal()) {
        Arco dest(destino);
        typename Lista<typename Grafo<C>::Arco>::Iterador posDest = posOrig.elementoActual().adyacentes.buscar(dest);

        if (!posDest.llegoAlFinal())
            return posDest.elementoActual().devolverCosto();
    }
}

// PRE CONDICION: existeVertice(origen) && existeVertice(destino)
// POST CONDICION: existeArco(origen, destino)
template <typename C> void Grafo<C>::agregarArco(int origen, int destino, const C & costo)
{
    assert(existeVertice(origen) && existeVertice(destino));

    Vertice orig(origen);
    typename Lista<Vertice>::Iterador posOrig = vertices.buscar(orig);

    if (!posOrig.llegoAlFinal()) {
        Arco dest(destino, costo);
        typename Lista<typename Grafo<C>::Arco>::Iterador posDest = posOrig.elementoActual().adyacentes.buscar(dest);

        if (!posDest.llegoAlFinal()) {
            posDest.elementoActual().devolverCosto() = costo;
        } else {
            Arco ady(destino, costo);
            posOrig.elementoActual().adyacentes.agregarFinal(ady);
        }
    }
}

// POST CONDICION: !existeArco(origen, destino)
template <typename C> void Grafo<C>::eliminarArco(int origen, int destino)
{
    Vertice orig(origen);
    typename Lista<Vertice>::Iterador posOrig = vertices.buscar(orig);

    if (!posOrig.llegoAlFinal()) {
        Arco dest(destino);
        posOrig.elementoActual().adyacentes.eliminar(posOrig.elementoActual().adyacentes.buscar(dest));
    }
}

template <typename C> void Grafo<C>::vaciar()
{
    vertices.vaciar();
}


/*
 * Se sobrecaga el operador de salida de los streams para permitir mostrar un grafo genrico.
 * Tener en cuenta que se asume que el tipo del costo soporta dicho operador.
 */

template <typename C> std::ostream & operator << (std::ostream & salida, const Grafo<C> & grafo)
{
	// Recorremos todos los vertices
	Lista<int> vertices;
	grafo.devolverVertices(vertices);
	Lista<int>::Iterador itVert = vertices.devolverIterador();
	while (!itVert.llegoAlFinal()) {
		Lista<typename Grafo<C>::Arco> adyacentes;
		grafo.devolverAdyacentes(itVert.elementoActual(), adyacentes);
        // La lista de arcos soporta el operador de salida de streams
		salida << "Vertice: " << itVert.elementoActual() << " Adyacentes: " << adyacentes;

		itVert.avanzar();
	}
	return salida;
}


/*
 * Funcin para cargar un grafo (con tipo de costo int) desde un archivo de texto.
 * El formato es:
 * - Cantidad de vrtices.
 * - Luego los identificadores de los vrtices, uno por lnea.
 * - Cantidad de arcos.
 * - Luego cada arco con el formato "origen,destino,costo" (sin espacios entre las comas), uno por lnea.
 */

void cargarGrafo(Grafo<int> & grafo, std::string ruta)
{
    grafo.vaciar();

    std::fstream archivo(ruta.c_str(), std::ios::in);

    if (archivo.is_open()) {
        if (!archivo.eof()) {
            int cantVertices = 0;
            archivo >> cantVertices;
            int indVertices = 0;
            while (indVertices < cantVertices) {
                int vertice = 0;
                archivo >> vertice;
                grafo.agregarVertice(vertice);
                indVertices++;
            }
            std::cout << "Se cargaron correctamente " << indVertices << " vertices.\n";

            if (!archivo.eof()) {
                int cantArcos = 0;
                archivo >> cantArcos;
                int indArcos = 0;
                while (indArcos < cantArcos) {
                    std::string linea;
                    archivo >> linea;

                    size_t pos1 = linea.find(",");
                    int origen = atoi(linea.substr(0, pos1).c_str());
                    size_t pos2 = linea.find(",", pos1+1);
                    int destino = atoi(linea.substr(pos1 + 1, pos2 - pos1).c_str());
                    int costo = atoi(linea.substr(pos2+1, linea.size() - pos2).c_str());

                    grafo.agregarArco(origen, destino, costo);
                    indArcos++;
                }
                std::cout << "Se cargaron correctamente "<< indArcos << " arcos.\n";
            }
        }
    } else {
        std::cout << "No se pudo abrir el archivo: " << ruta << "\n";
    }

    archivo.close();
}


#endif /* GRAFO_H_ */
